"""
Policy training patterns for use with `Evosax`_ neuro-evolution package

Classes definitions and functionality for training of evolutionary
policies (via `Evosax`_) using esquilax as a multi-agent
(and potentially multi-policy) training environment.

.. _Evosax: https://github.com/RobertTLange/evosax
"""
from . import tree_utils
from .strategy import BasicStrategy, Strategy
from .training import test, train
from .types import TrainingData



import typing
from functools import partial
from typing import Any

import chex
import jax

from esquilax import utils


def amap(f: typing.Callable) -> typing.Callable:
    """
    Decorator that maps a function over agent state data

    Maps an update function over state data,
    broadcasting model parameters, and assigning unique
    random keys. The size of the output data is the same
    as the inputs, i.e. it produces a result for each agent.

    Inputs and outputs can be PyTrees of arrays
    where each array entry represent an agent. See
    below for examples.

    Examples
    --------

    .. testsetup:: amap

       import esquilax
       import jax

    .. testcode:: amap

       @esquilax.transforms.amap
       def foo(k, p, x):
           return p + x

       k = jax.random.PRNGKey(101)
       a = jax.numpy.arange(5)

       foo(k, 2, a)
       # [2, 3, 4, 5, 6]

    .. doctest:: amap
       :hide:

       >>> foo(k, 2, a).tolist()
       [2, 3, 4, 5, 6]

    Arguments can also be PyTrees, and parameters
    ``None`` if unused

    .. testcode:: amap

       @esquilax.transforms.amap
       def foo(k, _, x):
           return x[0] + x[1]

       k = jax.random.PRNGKey(101)
       a = (jax.numpy.arange(5), jax.numpy.arange(5))

       foo(k, None, a)
       # [0, 2, 4, 6, 8]

    .. doctest:: amap
       :hide:

       >>> foo(k, None, a).tolist()
       [0, 2, 4, 6, 8]

    Arguments can also multidimensional (as long as
    the first axis is the number of agents), and outputs
    can be PyTrees

    .. testcode:: amap

       @esquilax.transforms.amap
       def foo(k, _, x):
           # Returns a tuple
           return x[0], x[1]

       k = jax.random.PRNGKey(101)
       a = jax.numpy.arange(10).reshape(5, 2)

       foo(k, None, a)
       # ([0, 2, 4, 6, 8], [1, 3, 5, 7, 9])

    .. doctest:: amap
       :hide:

       >> foo(k, None, a)
       ([0, 2, 4, 6, 8], [1, 3, 5, 7, 9])

    Parameters
    ----------
    f
        Update function that should have the signature

        .. code-block:: python

           def f(k, params, x, **static_kwargs):
               ...
               return y

        where the arguments are:

        - ``k``: A JAX PRNGKey
        - ``params``: Parameters (shared across the map)
        - ``x``: State data to map over.
        - ``**static_kwargs``: Any values required at compile
          time by JAX can be passed as keyword arguments.
    """

    keyword_args = utils.functions.get_keyword_args(f)

    @partial(jax.jit, static_argnames=keyword_args)
    def _self(k: chex.PRNGKey, params: Any, x: Any, **static_kwargs) -> Any:
        n = utils.functions.get_size(x)
        keys = jax.random.split(k, n)
        results = jax.vmap(partial(f, **static_kwargs), in_axes=(0, None, 0))(
            keys, params, x
        )
        return results

    return _self



"""
RL agent definitions
"""
from typing import Tuple

import chex
import jax
import jax.numpy as jnp
import optax
from flax import linen as nn
from flax.training.train_state import TrainState
from typing_extensions import Self


class Agent(TrainState):
    """
    Abstract rl-agent class

    Outlines functionality required for agents to used
    as part of esquilax rl training functionality.

    This class extends
    :py:class:`flax.training.train_state.TrainState`,
    containing a network function and parameters, along with an
    optimiser and corresponding state.

    An ``Agent`` can represent multiple agents, dependent
    on it's expected observation and returned action shapes.
    """

    def sample_actions(
        self, key: chex.PRNGKey, observations: chex.Array
    ) -> Tuple[chex.ArrayTree, chex.ArrayTree]:
        """
        Sample actions given observations

        Parameters
        ----------
        key
            JAX random key.
        observations
            Environment observations

        Returns
        -------
        tuple
            Tuple containing actions and
            any corresponding action values
        """
        raise NotImplementedError

    def update(self, key: chex.PRNGKey, trajectories) -> Tuple[Self, chex.ArrayTree]:
        """
        Update the state of the agent from observed trajectories

        Parameters
        ----------
        key
            JAX random key.
        trajectories
            Structure of observation-action environment update trajectories.

        Returns
        -------
        tuple[esquilax.ml.rl.Agent, chex.ArrayTree]
            Tuple containing an updated ``Agent`` and any associated data
            e.g. training losses from the update.
        """
        raise NotImplementedError

    def apply_grads(self, *, grads: chex.ArrayTree, **kwargs) -> Self:
        """
        Apply gradients to the agent parameters and optimiser

        Parameters
        ----------
        grads
            Gradients corresponding to the agent parameters.
        **kwargs
            Any keyword arguments to pass to underlying ``apply_gradients`` function.

        Returns
        -------
        esquilax.ml.rl.Agent
            Updated agent
        """
        raise NotImplementedError


class SharedPolicyAgent(Agent):
    """
    Agent with a single trained policy
    """

    @classmethod
    def init(
        cls,
        key: chex.PRNGKey,
        model: nn.Module,
        tx: optax.GradientTransformation,
        observation_shape: Tuple[int, ...],
    ) -> Self:
        """
        Initialise the agent from a network

        Parameters
        ----------
        key
            JAX random key.
        model
            Flax neural network model definition.
        tx
            Optax optimiser.
        observation_shape
            Shape of observations

        Returns
        -------
        esquilax.ml.rl.SharedPolicyAgent
            Initialised agent.
        """
        fake_args = jnp.zeros(observation_shape)
        params = model.init(key, fake_args)
        return cls.create(apply_fn=model.apply, params=params, tx=tx)

    def apply_grads(self, *, grads, **kwargs) -> Self:
        """
        Apply gradients to the agent parameters and optimiser

        Parameters
        ----------
        grads
            Gradients corresponding to the agent parameters.
        **kwargs
            Any keyword arguments to pass to underlying ``apply_gradients`` function.

        Returns
        -------
        esquilax.ml.rl.SharedPolicyAgent
            Updated agent
        """
        return self.apply_gradients(grads=grads, **kwargs)


class BatchPolicyAgent(Agent):
    """
    Agent with parameters specific to each agent

    Agent intended for use where a group of agents is each
    trained on its own individual policy (but they share the
    same network structure, and optimiser type).
    """

    @classmethod
    def init(
        cls,
        key: chex.PRNGKey,
        model: nn.Module,
        tx: optax.GradientTransformation,
        observation_shape: Tuple[int, ...],
        n_agents: int,
    ) -> Self:
        """
        Initialise the agent from a network

        Initialises individual parameters and optimisers
        for required agents.

        Parameters
        ----------
        key
            JAX random key.
        model
            Flax neural network model definition.
        tx
            Optax optimiser.
        observation_shape
            Shape of observations.
        n_agents
            Number of agents to initialise state for.

        Returns
        -------
        esquilax.ml.rl.BatchPolicyAgent
            Initialised agent.
        """
        fake_args = jnp.zeros(observation_shape)

        def init(_k):
            params = model.init(_k, fake_args)
            return cls.create(apply_fn=model.apply, params=params, tx=tx)

        keys = jax.random.split(key, n_agents)
        return jax.vmap(init)(keys)

    def apply_grads(self, *, grads, **kwargs) -> Self:
        """
        Apply gradients to the agent parameters and optimiser

        Apply gradients across the set of agents, with each
        agent updated individually.

        Parameters
        ----------
        grads
            Gradients corresponding to the agent parameters. Each agent
            should receive its own set of gradients, i.e. gradients
            should have a shape ``[n-agents, n-parameters, ...]``.
        **kwargs
            Any keyword arguments to pass to underlying ``apply_gradients`` function.

        Returns
        -------
        esquilax.ml.rl.BatchPolicyAgent
            Updated agent
        """
        return jax.vmap(lambda a, g: a.apply_gradients(grads=g, **kwargs))(self, grads)



"""
RL and Neuro-evolution functionality and utilities

Functions and utilities for multi-agent reinforcement-learning
and neuro-evolution training, where Esquilax simulations
plays the role of training environments.

Training supports several configurations, including multiple RL agents
with individual policies, and training of multiple agent
types or strategies inside the same environment.
"""
from . import evo, rl
from .updates import get_actions, sample_actions



"""
Simulation and experiment runners
"""
from functools import partial
from typing import Any, Callable, Tuple, Union

import chex
import jax
import jax_tqdm


def sim_runner(
    step_fun: Callable,
    params: Any,
    initial_state: chex.ArrayTree,
    n_steps: int,
    rng: Union[chex.PRNGKey, int],
    show_progress: bool = True,
    **static_kwargs,
) -> Tuple[Any, Any, chex.PRNGKey]:
    """
    Run a simulation and track state

    Repeated applies a provided simulation update function and
    records simulation state over the course of execution.

    Parameters
    ----------
    step_fun: Callable
        Update function that should have the signature

        .. code-block:: python

           def step(i, k, params, state, **static_kwargs):
               ...
               return new_state, records

        where the arguments are

        - ``i``: The current step number
        - ``k``: A JAX random key
        - ``params``: Sim parameters
        - ``state``: The current simulation state
        - ``**static_kwargs``: Static keyword arguments

        and returns

        - ``new_state``: Updated simulation state
        - ``records``: State data to be recorded

    params
        Simulation parameters. Parameters are constant over the
        course of the simulation.
    initial_state
        Initial simulation state.
    n_steps
        Number of steps to run.
    rng
        Either an integer random seed, or a JAX PRNGKey.
    show_progress
        If ``True`` a progress bar will be shown.
    **static_kwargs
        Any keyword static values passed to the step function.
        These should be used for any values or functionality required
        to be known at compile time by JAX.

    Returns
    -------
    [Any, Any, chex.PRNGKey]
        Tuple containing

        - The final state of the simulation
        - Recorded values
        - Update random key
    """

    key = rng if isinstance(rng, chex.PRNGKey) else jax.random.PRNGKey(rng)
    step_fun = partial(step_fun, **static_kwargs)

    def step(carry, i):
        k, state = carry
        k, step_key = jax.random.split(k)
        new_state, records = step_fun(i, step_key, params, state)
        return (k, new_state), records

    if show_progress:
        step = jax_tqdm.scan_tqdm(n_steps, desc="Step")(step)

    (key, final_state), record_history = jax.lax.scan(
        step, (key, initial_state), jax.numpy.arange(n_steps)
    )

    return final_state, record_history, key



"""
Reinforcement learning utilities and training functionality
"""
from .agents import Agent, BatchPolicyAgent, SharedPolicyAgent
from .environment import Environment
from .training import test, train
from .types import Trajectory



"""
RL training and testing functionality
"""

from typing import Callable, Tuple

import chex
import jax
import jax.numpy as jnp
import jax_tqdm

from esquilax.typing import TEnvParams, TypedPyTree

from . import tree_utils
from .agents import Agent
from .environment import Environment
from .types import Trajectory


def _step(env_params: TEnvParams, env: Environment) -> Callable:
    def inner_step(carry, _) -> Tuple[Tuple, Trajectory]:
        _k, _env_state, _obs, _agents = carry
        _k, _k_act, _k_step = jax.random.split(_k, 3)
        _actions, _action_values = tree_utils.sample_actions(_k_act, _agents, _obs)
        _new_obs, _env_state, _rewards, _done = env.step(
            _k_step, env_params, _env_state, _actions
        )
        trajectories = jax.tree.map(
            lambda *x: Trajectory(
                obs=x[0],
                actions=x[1],
                action_values=x[2],
                rewards=x[3],
                done=x[4],
            ),
            _new_obs,
            _actions,
            _action_values,
            _rewards,
            _done,
        )
        return (
            (_k, _env_state, _new_obs, _agents),
            trajectories,
        )

    return inner_step


def train(
    key: chex.PRNGKey,
    agents: TypedPyTree[Agent],
    env: Environment,
    env_params: TEnvParams,
    n_epochs: int,
    n_env: int,
    n_env_steps: int,
    show_progress: bool = True,
) -> Tuple[TypedPyTree[Agent], chex.ArrayTree, chex.ArrayTree]:
    """
    Train an RL-agent or agents with a given environment

    Parameters
    ----------
    key
        JAX random key.
    agents
        RL agent, or collection of agents. Multiple
        agents/policies can be provided to allow for
        training of multiple agent types.
    env
        Training environment/simulation. This should
        implement the :py:class:`esquilax.ml.rl.Environment`
        interface.
    env_params
        Environment parameters.
    n_epochs
        Number of training epochs.
    n_env
        Number of environments to train across per epoch.
    n_env_steps
        Number of steps to run in each environment per epoch.
    show_progress
        If ``True`` a training progress bar will be displayed.
        Default ``True``.

    Returns
    -------
    tuple
        Tuple containing:

        - Update agent or collection of agents
        - Training rewards
        - Additional training data
    """
    step = _step(env_params, env)

    def sample_trajectories(_k, _agents):
        _obs, _state = env.reset(_k, env_params)
        _, trajectories = jax.lax.scan(
            step, (_k, _state, _obs, _agents), None, length=n_env_steps
        )
        return trajectories

    def epoch(carry, _):
        _k, _agents = carry
        _k, _k_sample, _k_train = jax.random.split(_k, 3)
        _k_sample = jax.random.split(_k_sample, n_env)
        _trajectories = jax.vmap(sample_trajectories, in_axes=(0, None))(
            _k_sample, _agents
        )
        _agents, _train_data = tree_utils.update_agents(
            _k_train, _agents, _trajectories
        )

        return (_k, _agents), (
            jax.tree.map(
                lambda _, t: t.rewards,
                _agents,
                _trajectories,
                is_leaf=lambda x: isinstance(x, Agent),
            ),
            _train_data,
        )

    if show_progress:
        epoch = jax_tqdm.scan_tqdm(n_epochs, desc="Epoch")(epoch)

    (_, agents), (rewards, train_data) = jax.lax.scan(
        epoch,
        (key, agents),
        jnp.arange(n_epochs),
    )

    return agents, rewards, train_data


def test(
    key: chex.PRNGKey,
    agents: TypedPyTree[Agent],
    env: Environment,
    env_params: TEnvParams,
    n_env: int,
    n_env_steps: int,
    show_progress: bool = True,
) -> Trajectory:
    """
    Test agent(s) performance

    Asses agents against a test environment,
    returning trajectory data gathered over training.

    Parameters
    ----------
    key
        JAX random key.
    agents
        RL agent, or collection of agents. Multiple
        agents/policies can be provided to allow for
        testing of multiple agent types.
    env
        Training environment/simulation. This should
        implement a Gymnax Environment base class.
    env_params
        Environment parameters.
    n_env
        Number of environments to test across.
    n_env_steps
        Number of steps to run in each environment.
    show_progress
        If ``True`` a testing progress bar will be displayed.
        Default ``True``

    Returns
    -------
    esquilax.ml.rl.Trajectory
        Update trajectories gathered over testing.
    """
    step = _step(env_params, env)

    if show_progress:
        step = jax_tqdm.scan_tqdm(n_env_steps, desc="Step")(step)

    def sample_trajectories(_k, _agents):
        _obs, _state = env.reset(_k, env_params)
        _, _trajectories = jax.lax.scan(
            step, (_k, _state, _obs, _agents), None, length=n_env_steps
        )
        return _trajectories

    k_sample = jax.random.split(key, n_env)

    trajectories = jax.vmap(sample_trajectories, in_axes=(0, None))(k_sample, agents)

    return trajectories



"""
Wrappers for `Evosax`_ strategy functionality

Class definitions wrapping strategy functionality
for use when training policies within an Esquilax
simulation. Classes implementing
:py:class:`esquilax.ml.evo.Strategy` can be passed
to :py:class:`esquilax.ml.evo.train` which will then
automatically handle optimisation.

Multiple strategies can be passed to
:py:class:`esquilax.ml.evo.train` allowing training
of multiple strategies within the same class.

.. _Evosax: https://github.com/RobertTLange/evosax
"""
from functools import partial
from typing import Any, Dict, Tuple, Union

import chex
import evosax
import jax
from flax.typing import FrozenVariableDict


class Strategy:
    """
    Abstract class wrapping Evosax strategy functionality
    """

    def initialize(
        self, key: chex.PRNGKey, evo_params: evosax.EvoParams
    ) -> evosax.EvoState:
        """
        Initialise strategy state

        Parameters
        ----------
        key
            JAX random key
        evo_params: evosax.EvoParams
            Strategy parameters

        Returns
        -------
        evosax.EvoState
            Strategy state
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0,))
    def default_params(self) -> evosax.EvoParams:
        """
        Get default strategy parameters

        Returns
        -------
        evosax.EvoParams
            Strategy parameters
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0,))
    def reshape_params(self, population) -> chex.ArrayTree:
        """
        Reshape parameters

        Parameters
        ----------
        population
            Array of parameters sampled from the strategy.

        Returns
        -------
        Reshaped parameters
        """
        return population

    @partial(jax.jit, static_argnums=(0,))
    def shape_rewards(self, population, fitness: chex.ArrayTree) -> chex.ArrayTree:
        """
        Reshape environment rewards

        Rescale/reshape rewards generated during a simulation.

        .. warning::

           Evosax expects to fitness minimised, so either the
           training environment should return values to
           be minimised, or rewards returned from the
           environment should be rescaled by this method.

        Parameters
        ----------
        population
            Strategy population
        fitness
            Fitness/rewards

        Returns
        -------
        chex.ArrayTree
            Rescaled fitness
        """
        return fitness

    @partial(jax.jit, static_argnums=(0,))
    def ask(
        self,
        key: chex.PRNGKey,
        evo_state: evosax.EvoState,
        evo_params: evosax.EvoParams,
    ) -> Tuple[chex.ArrayTree, evosax.EvoState]:
        """
        Sample parameters from the current strategy state

        Parameters
        ----------
        key
            JAX random key
        evo_state
            Strategy state
        evo_params
            Strategy parameters

        Returns
        -------
        tuple
            Population array and updated state of the strategy
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0,))
    def tell(
        self,
        population: chex.ArrayTree,
        fitness: chex.ArrayTree,
        evo_state: evosax.EvoState,
        evo_params: evosax.EvoParams,
    ) -> evosax.EvoState:
        """
        Update strategy state

        Parameters
        ----------
        population
            Population array
        fitness
            Fitness array
        evo_state
            Strategy state
        evo_params
            Strategy params

        Returns
        -------
        evosax.EvoState
            Update state
        """
        raise NotImplementedError


class BasicStrategy(Strategy):
    """
    Basic strategy derived from a Flax neural network

    Wrapper around a strategy, with parameters
    initialised from a Flax neural-network.
    """

    param_reshaper: evosax.ParameterReshaper
    strategy: evosax.Strategy
    fitness_shaper: evosax.FitnessShaper

    def __init__(
        self,
        network_params: Union[FrozenVariableDict, Dict[str, Any]],
        strategy: evosax.Strategy,
        pop_size: int,
        centered_rank_fitness: bool = True,
        z_score_fitness: bool = False,
        maximize_fitness: bool = True,
        **strategy_kwargs,
    ):
        """
        Parameters
        ----------
        network_params
            Flax network parameters.
        strategy
            Evosax strategy class.
        pop_size
            Strategy population size.
        centered_rank_fitness
            Use ``centered_rank_fitness`` for fitness-shaping, default ``True``.
        z_score_fitness
            Use ``z_score_fitness`` for fitness-shaping, default ``False``.
        maximize_fitness
            Use ``maximize_fitness`` for fitness-shaping, default ``True``.

            .. warning::

               Evosax expects that fitness should be minimised, so this should
               be ``True`` if the environment returns rewards to be maximised.

        **strategy_kwargs
            Keyword arguments to pass to the strategy constructor.
        """
        self.param_reshaper = evosax.ParameterReshaper(network_params)
        self.strategy = strategy(
            popsize=pop_size,
            num_dims=self.param_reshaper.total_params,
            **strategy_kwargs,
        )
        self.fitness_shaper = evosax.FitnessShaper(
            centered_rank=centered_rank_fitness,
            z_score=z_score_fitness,
            maximize=maximize_fitness,
        )

    def initialize(
        self, key: chex.PRNGKey, evo_params: evosax.EvoParams
    ) -> evosax.EvoState:
        """
        Initialise strategy state

        Parameters
        ----------
        key
            JAX random key
        evo_params
            Strategy parameters

        Returns
        -------
        evosax.EvoState
            Strategy state
        """
        return self.strategy.initialize(key, evo_params)

    @partial(jax.jit, static_argnums=(0,))
    def default_params(self) -> evosax.EvoParams:
        """
        Get default strategy parameters

        Returns
        -------
        evosax.EvoParams
            Strategy parameters
        """
        return self.strategy.default_params

    def reshape_params(self, population) -> chex.ArrayTree:
        """
        Reshape parameters

        Parameters
        ----------
        population
            Array of parameters sampled from the strategy.

        Returns
        -------
        jax.numpy.array
            Rescaled parameters
        """
        return self.param_reshaper.reshape(population)

    def ask(
        self,
        key: chex.PRNGKey,
        evo_state: evosax.EvoState,
        evo_params: evosax.EvoParams,
    ) -> Tuple[chex.ArrayTree, evosax.EvoState]:
        """
        Sample parameters from the current strategy state

        Parameters
        ----------
        key
            JAX random key
        evo_state
            Strategy state
        evo_params
            Strategy parameters

        Returns
        -------
        tuple[jax.numpy.array, evosax.EvoState]
            Population array and updated state of the strategy
        """
        return self.strategy.ask_strategy(key, evo_state, evo_params)

    def shape_rewards(self, population, fitness: chex.ArrayTree) -> chex.ArrayTree:
        """
        Reshape rewards

        Rescale/reshape rewards generated during a simulation.

        Parameters
        ----------
        population
            Strategy population or collection of populations
        fitness
            Fitness/rewards (or collection of)

        Returns
        -------
        chex.ArrayTree
            Rescaled fitness
        """
        return self.fitness_shaper.apply(population, fitness)

    def tell(
        self,
        population: chex.ArrayTree,
        fitness: chex.ArrayTree,
        evo_state: evosax.EvoState,
        evo_params: evosax.EvoParams,
    ) -> evosax.EvoState:
        """
        Update strategy state

        Parameters
        ----------
        population
            Population array
        fitness
            Fitness array
        evo_state
            Strategy state
        evo_params
            Strategy params

        Returns
        -------
        evosax.EvoState
            Update state
        """
        return self.strategy.tell(population, fitness, evo_state, evo_params)



"""
Generic types
"""
from typing import Collection, TypeVar

TSimState = TypeVar("TSimState")
"""Generic simulation state"""
TSimParams = TypeVar("TSimParams")
"""Generic simulation parameters"""
TEnvState = TypeVar("TEnvState")
"""Generic environment state"""
TEnvParams = TypeVar("TEnvParams")
"""Generic environment parameters"""
T = TypeVar("T")
TypedPyTree = T | Collection[T]
"""PyTree with leaves of a single type"""



"""
Function and data utilities
"""
import inspect
from typing import Callable, List

import chex
import jax


def get_size(tree: chex.ArrayTree) -> int:
    """
    Get the length of arrays in a PyTree

    Also checks that arrays are all the same length

    Examples
    --------

    .. testsetup:: get_size

       import jax.numpy as jnp
       from esquilax.utils.functions import get_size

    .. doctest:: get_size

       >>> x = (jnp.arange(10), jnp.arange(10))
       >>> get_size(x)
       10

    Parameters
    ----------
    tree
        PyTree of arrays

    Raises
    ------
    AssertionError
        Raised if arrays do not have the same length.

    Returns
    -------
    int
        Length of arrays in the tree
    """
    lens = jax.tree_util.tree_map(lambda x: x.shape[0], tree)
    lens = jax.tree_util.tree_flatten(lens)[0]
    assert all([lens[0] == n for n in lens[1:]])
    return lens[0]


def get_keyword_args(f: Callable) -> List[str]:
    """
    Get keyword-only argument names from a function signature

    Examples
    --------

    .. testsetup:: get_keyword_args

       import jax.numpy as jnp
       from esquilax.utils.functions import get_keyword_args

    .. doctest:: get_keyword_args

       >>> def foo(a, b, *, c, d): pass
       >>> get_keyword_args(foo)
       ['c', 'd']

    Parameters
    ----------
    f
        Function to inspect

    Returns
    -------
    list[str]
        List of keyword-only argument names.
    """
    sig = inspect.signature(f)
    key_word_args = [
        k for k, p in sig.parameters.items() if p.kind == inspect.Parameter.KEYWORD_ONLY
    ]
    return key_word_args



from typing import Any, Optional, Tuple, Type

import chex
import jax

from esquilax.typing import TypedPyTree


def key_tree_split(
    key: chex.PRNGKey, tree: TypedPyTree[Any], typ: Optional[Type] = None
) -> TypedPyTree[chex.PRNGKey]:
    """
    Generate random keys for PyTree leaves

    Split a random key, generating a PyTree of keys matching
    the structure of the provided tree. A type can be
    provided to define an expected type for leaves of the
    tree.

    Parameters
    ----------
    key
        JAX random key.
    tree
        Pytree to map over.
    typ
        If provided, leaves of the tree
        will be those that match the provided type.

    Returns
    -------
    esquilax.typing.TypedPyTree[chex.PRNGKey]
        PyTree with random keys as leaves and the
        same structure as the argument tree
    """
    if typ is None:
        is_leaf = None
    else:

        def is_leaf(x):
            return isinstance(x, typ)

    tree_def = jax.tree.structure(tree, is_leaf=is_leaf)
    keys = jax.random.split(key, tree_def.num_leaves)
    keys = jax.tree.unflatten(tree_def, keys)
    return keys


def transpose_tree_of_tuples(
    tree_a: TypedPyTree[Any], tree_b, n: int, typ: Optional[Type] = None
) -> Tuple:
    """
    Transpose a tree containing tuples, into a tuple of the outer tree

    Parameters
    ----------
    tree_a
        Tree to take outer structure from.
    tree_b
        Tree to transpose
    n
        Number of values to unpack
    typ
        Optional type to use as leaf check in ``tree_a``.
    Returns
    -------
    tuple
        Tuple of PyTrees
    """
    if typ is None:
        is_leaf = None
    else:

        def is_leaf(x):
            return isinstance(x, typ)

    return tuple(
        jax.tree.map(
            lambda _, x: x[i],
            tree_a,
            tree_b,
            is_leaf=is_leaf,
        )
        for i in range(n)
    )



"""
Data types for neuro-evolution training
"""
import chex


@chex.dataclass
class TrainingData:
    """
    Dataclass tracking training rewards and recorded data

    Simulations used for training should return
    a ``TrainingData`` class as its recorded data.

    Examples
    --------

    .. code-block:: python

       def step(self, i, k, params, state):
           ...
           # Rewards are returned as part of the TrainingData
           return (
               state,
               TrainingData(rewards=rewards, records=records)
           )

    Parameters
    ----------
    rewards
        PyTree of rewards generated by simulation agents.
    records
        Any additional data recorded over the course
        of the simulation.
    """

    rewards: chex.Array
    records: chex.ArrayTree



"""
Agent PyTree mapping utilities
"""
from typing import Tuple

import chex
import jax

from esquilax.ml import common
from esquilax.typing import TypedPyTree

from .agents import Agent
from .types import Trajectory


def sample_actions(
    key: chex.PRNGKey, agents: TypedPyTree[Agent], observations: chex.ArrayTree
) -> Tuple[chex.ArrayTree, chex.ArrayTree]:
    """
    Map over a tree of agents, sampling actions from observations

    Can be used to sample actions across multiple
    RL policies.

    Parameters
    ----------
    key
        JAX random key.
    agents
        Pytree of RL Agents.
    observations
        PyTree of observations, with tree structure corresponding
        to the agents.

    Returns
    -------
    tuple[chex.ArrayTree, chex.ArrayTree]
        Tuple containing actions, and any additional values
        associated with the actions. Both have the same
        tree structure as the argument agents.
    """
    keys = common.key_tree_split(key, agents, typ=Agent)
    results = jax.tree.map(
        lambda agent, k, obs: agent.sample_actions(k, obs),
        agents,
        keys,
        observations,
        is_leaf=lambda x: isinstance(x, Agent),
    )
    actions, action_values = common.transpose_tree_of_tuples(agents, results, 2, Agent)
    return actions, action_values


def update_agents(
    key: chex.PRNGKey,
    agents: TypedPyTree[Agent],
    trajectories: TypedPyTree[Trajectory],
) -> Tuple[TypedPyTree[Agent], chex.ArrayTree]:
    """
    Update agent states from gathered trajectories

    Parameters
    ----------
    key
        JAX random key.
    agents
        PyTree of RL-agents.
    trajectories
        PyTree of environment trajectories.

    Returns
    -------
    tuple[esquilax.typing.TypedPyTree[esquilax.ml.rl.Agent], chex.ArrayTree]
        Tuple containing PyTrees of updated
        agents, and any data returned from training
        (e.g. training loss). Both trees have the same
        structure as the argument agents.
    """
    keys = common.key_tree_split(key, agents, typ=Agent)

    updates = jax.tree.map(
        lambda agent, k, t: agent.update(k, t),
        agents,
        keys,
        trajectories,
        is_leaf=lambda x: isinstance(x, Agent),
    )
    agents, train_data = common.transpose_tree_of_tuples(agents, updates, 2, Agent)
    return agents, train_data
